/*
 * JzRISC LCD controller
 *
 * Copyright (C) Xiangfu Liu <xiangfu@sharism.cc>
 * Copyright (C) 2015, 2016, 2017 Paul Boddie <paul@boddie.org.uk>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of
 * the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.	 See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor,
 * Boston, MA  02110-1301, USA
 */

#include "sdram.h"
#include "jzlcd.h"
#include "cpu.h"
#include "board.h"

#define align2(n) (n)=((((n)+1)>>1)<<1)
#define align4(n) (n)=((((n)+3)>>2)<<2)
#define align8(n) (n)=((((n)+7)>>3)<<3)

extern vidinfo_t panel_info;

static uint16_t lcd_get_panels(vidinfo_t *vid)
{
	struct jzfb_info *jzfb = vid->jz_fb;
	return ((jzfb->cfg & MODE_MASK) == MODE_STN_MONO_DUAL) ||
		((jzfb->cfg & MODE_MASK) == MODE_STN_COLOR_DUAL) ? 2 : 1;
}



/* Functions returning region sizes. */

static uint32_t lcd_get_size(vidinfo_t *vid)
{
	/* Lines must be aligned to a word boundary. */
	uint32_t line_length = ALIGN((vid->jz_fb->w * vid->jz_fb->bpp) / 8, sizeof(uint32_t));
        return line_length * vid->jz_fb->h;
}

static uint32_t lcd_get_aligned_size(vidinfo_t *vid)
{
	/* LCD_CTRL_BST_16 requires 16-word alignment. */
        return ALIGN(lcd_get_size(vid), 16 * sizeof(uint32_t));
}

static uint32_t lcd_get_min_size(vidinfo_t *vid)
{
	/* Lines must be aligned to a word boundary. */
	uint32_t line_length = ALIGN((vid->jz_fb->w * 32) / 8, sizeof(uint32_t));
        return line_length * vid->jz_fb->h;
}

static uint32_t lcd_get_aligned_min_size(vidinfo_t *vid)
{
	/* LCD_CTRL_BST_16 requires 16-word alignment. */
        return ALIGN(lcd_get_min_size(vid), 16 * sizeof(uint32_t));
}

static uint32_t lcd_get_palette_size(vidinfo_t *vid)
{
	if (vid->jz_fb->bpp < 12)
		return (1 << (vid->jz_fb->bpp)) * sizeof(uint16_t);
	else
		return 0;
}

static uint32_t lcd_get_aligned_palette_size(vidinfo_t *vid)
{
	/* LCD_CTRL_BST_16 requires 16-word alignment. */
	return ALIGN(lcd_get_palette_size(vid), 16 * sizeof(uint32_t));
}

static uint32_t lcd_get_descriptors_size()
{
	return 3 * sizeof(struct jz_fb_dma_descriptor);
}

static uint32_t lcd_get_total_size(vidinfo_t *vid)
{
	uint32_t size = lcd_get_aligned_size(vid) * lcd_get_panels(vid);
	uint32_t min_size = lcd_get_aligned_min_size(vid);

        /* Round up to nearest full page, or MMU section if defined. */
	return ALIGN((size >= min_size ? size : min_size) + lcd_get_aligned_palette_size(vid) + lcd_get_descriptors_size(), PAGE_SIZE);
}



/* Functions returning addresses of each data region. */

static uint32_t lcd_get_palette(uint32_t addr, vidinfo_t *vid)
{
        /* Allocate memory at the end of the region for the palette. */
	return addr - lcd_get_aligned_palette_size(vid);
}

static uint32_t lcd_get_descriptors(uint32_t addr, vidinfo_t *vid)
{
        /* Allocate memory before the palette for the descriptor array. */
	return lcd_get_palette(addr, vid) - lcd_get_descriptors_size();
}

static uint32_t lcd_get_framebuffer(uint32_t addr, uint16_t panel, vidinfo_t *vid)
{
        /* Allocate pages for the frame buffer and palette. */
        return addr - lcd_get_total_size(vid) + (panel * lcd_get_aligned_size(vid));
}



/* Initialisation functions. */

static void jz_lcd_desc_init(vidinfo_t *vid)
{
	struct jz_fb_dma_descriptor *descriptors;
	struct jz_mem_info * fbi;

	fbi = &vid->jz_mem;

	/* Allocate space for descriptors before the palette entries. */

	descriptors = (struct jz_fb_dma_descriptor *) lcd_get_descriptors(get_memory_size(), vid);
	fbi->dmadesc_fblow = (struct jz_fb_dma_descriptor *) &descriptors[0];
	fbi->dmadesc_fbhigh = (struct jz_fb_dma_descriptor *) &descriptors[1];
	fbi->dmadesc_palette = (struct jz_fb_dma_descriptor *) &descriptors[2];

	/* Populate descriptors. */

	if (lcd_get_panels(vid) == 2)
	{
		fbi->dmadesc_fblow->fdadr = fbi->dmadesc_fblow;
		fbi->dmadesc_fblow->fsadr = lcd_get_framebuffer(get_memory_size(), 1, vid);
		fbi->dmadesc_fblow->fidr  = 0;
		fbi->dmadesc_fblow->ldcmd = lcd_get_size(vid) / 4 ;

		fbi->fdadr1 = fbi->dmadesc_fblow; /* only used in dual-panel mode */
	}

	fbi->dmadesc_fbhigh->fsadr = fbi->screen;
	fbi->dmadesc_fbhigh->fidr = 0;
	fbi->dmadesc_fbhigh->ldcmd =  lcd_get_size(vid) / 4; /* length in words */

	if (vid->jz_fb->bpp < 12)
	{
		fbi->dmadesc_palette->fsadr = fbi->palette;
		fbi->dmadesc_palette->fidr  = 0;
		fbi->dmadesc_palette->ldcmd = (lcd_get_palette_size(vid) / 4) | (1<<28);

		/* assume any mode with <12 bpp is palette driven */
		fbi->dmadesc_palette->fdadr = fbi->dmadesc_fbhigh;
		fbi->dmadesc_fbhigh->fdadr = fbi->dmadesc_palette;
		/* flips back and forth between pal and fbhigh */
		fbi->fdadr0 = fbi->dmadesc_palette;
	} else {
		/* palette shouldn't be loaded in true-color mode */
		fbi->dmadesc_fbhigh->fdadr = fbi->dmadesc_fbhigh;
		fbi->fdadr0 = fbi->dmadesc_fbhigh; /* no pal just fbhigh */
	}

	flush_cache_all();
}

static uint32_t jz_lcd_stn_init(uint32_t stnH, vidinfo_t *vid)
{
	struct jzfb_info *jzfb = vid->jz_fb;
	uint32_t val = 0;

	switch (jzfb->bpp) {
	case 1:
		/* val |= LCD_CTRL_PEDN; */
	case 2:
		val |= LCD_CTRL_FRC_2;
		break;
	case 4:
		val |= LCD_CTRL_FRC_4;
		break;
	case 8:
	default:
		val |= LCD_CTRL_FRC_16;
		break;
	}

	switch (jzfb->cfg & STN_DAT_PINMASK) {
	case STN_DAT_PIN1:
		/* Do not adjust the hori-param value. */
		break;
	case STN_DAT_PIN2:
		align2(jzfb->hsw);
		align2(jzfb->elw);
		align2(jzfb->blw);
		break;
	case STN_DAT_PIN4:
		align4(jzfb->hsw);
		align4(jzfb->elw);
		align4(jzfb->blw);
		break;
	case STN_DAT_PIN8:
		align8(jzfb->hsw);
		align8(jzfb->elw);
		align8(jzfb->blw);
		break;
	}

	REG_LCD_VSYNC = (0 << 16) | jzfb->vsw;
	REG_LCD_HSYNC = ((jzfb->blw+jzfb->w) << 16) | (jzfb->blw+jzfb->w+jzfb->hsw);

	/* Screen setting */
	REG_LCD_VAT = ((jzfb->blw + jzfb->w + jzfb->hsw + jzfb->elw) << 16) | (stnH + jzfb->vsw + jzfb->bfw + jzfb->efw);
	REG_LCD_DAH = (jzfb->blw << 16) | (jzfb->blw + jzfb->w);
	REG_LCD_DAV = (0 << 16) | (stnH);

	/* AC BIAs signal */
	REG_LCD_PS = (0 << 16) | (stnH+jzfb->vsw+jzfb->efw+jzfb->bfw);

	return val;
}

static void jz_lcd_tft_init(vidinfo_t *vid)
{
	struct jzfb_info *jzfb = vid->jz_fb;
	REG_LCD_VSYNC = (0 << 16) | jzfb->vsw;
	REG_LCD_HSYNC = (0 << 16) | jzfb->hsw;
	REG_LCD_DAV =((jzfb->vsw+jzfb->bfw) << 16) | (jzfb->vsw +jzfb->bfw+jzfb->h);
	REG_LCD_DAH = ((jzfb->hsw + jzfb->blw) << 16) | (jzfb->hsw + jzfb->blw + jzfb->w );
	REG_LCD_VAT = (((jzfb->blw + jzfb->w + jzfb->elw + jzfb->hsw)) << 16) \
		| (jzfb->vsw + jzfb->bfw + jzfb->h + jzfb->efw);
}

static void jz_lcd_samsung_init(uint32_t pclk, vidinfo_t *vid)
{
	struct jzfb_info *jzfb = vid->jz_fb;
	uint32_t total, tp_s, tp_e, ckv_s, ckv_e;
	uint32_t rev_s, rev_e, inv_s, inv_e;

	jz_lcd_tft_init(vid);

	total = jzfb->blw + jzfb->w + jzfb->elw + jzfb->hsw;
	tp_s = jzfb->blw + jzfb->w + 1;
	tp_e = tp_s + 1;
	/* ckv_s = tp_s - jz_clocks.pixclk/(1000000000/4100); */
	ckv_s = tp_s - pclk/(1000000000/4100);
	ckv_e = tp_s + total;
	rev_s = tp_s - 11;	/* -11.5 clk */
	rev_e = rev_s + total;
	inv_s = tp_s;
	inv_e = inv_s + total;
	REG_LCD_CLS = (tp_s << 16) | tp_e;
	REG_LCD_PS = (ckv_s << 16) | ckv_e;
	REG_LCD_SPL = (rev_s << 16) | rev_e;
	REG_LCD_REV = (inv_s << 16) | inv_e;
	jzfb->cfg |= STFT_REVHI | STFT_SPLHI;
}

static void jz_lcd_sharp_init(vidinfo_t *vid)
{
	struct jzfb_info *jzfb = vid->jz_fb;
	uint32_t total, cls_s, cls_e, ps_s, ps_e;
	uint32_t spl_s, spl_e, rev_s, rev_e;

	jz_lcd_tft_init(vid);

	total = jzfb->blw + jzfb->w + jzfb->elw + jzfb->hsw;
	spl_s = 1;
	spl_e = spl_s + 1;
	cls_s = 0;
	cls_e = total - 60;	/* > 4us (pclk = 80ns) */
	ps_s = cls_s;
	ps_e = cls_e;
	rev_s = total - 40;	/* > 3us (pclk = 80ns) */
	rev_e = rev_s + total;
	jzfb->cfg |= STFT_PSHI; 
	REG_LCD_SPL = (spl_s << 16) | spl_e;
	REG_LCD_CLS = (cls_s << 16) | cls_e;
	REG_LCD_PS = (ps_s << 16) | ps_e;
	REG_LCD_REV = (rev_s << 16) | rev_e;
}

static uint32_t jz_lcd_get_pixel_clock(vidinfo_t *vid)
{
	struct jzfb_info *jzfb = vid->jz_fb;
	uint32_t pclk;

	/* Derive pixel clock from frame clock. */

	if ( (jzfb->cfg & MODE_MASK) != MODE_8BIT_SERIAL_TFT) {
		pclk = jzfb->fclk * (jzfb->w + jzfb->hsw + jzfb->elw + jzfb->blw) *
			(jzfb->h + jzfb->vsw + jzfb->efw + jzfb->bfw);
	} else {
		/* serial mode: Hsync period = 3*Width_Pixel */
		pclk = jzfb->fclk * (jzfb->w*3 + jzfb->hsw + jzfb->elw + jzfb->blw) *
			(jzfb->h + jzfb->vsw + jzfb->efw + jzfb->bfw);
	}

	if (((jzfb->cfg & MODE_MASK) == MODE_STN_COLOR_SINGLE) ||
	    ((jzfb->cfg & MODE_MASK) == MODE_STN_COLOR_DUAL))
		pclk = (pclk * 3);

	if (((jzfb->cfg & MODE_MASK) == MODE_STN_COLOR_SINGLE) ||
	    ((jzfb->cfg & MODE_MASK) == MODE_STN_COLOR_DUAL) ||
	    ((jzfb->cfg & MODE_MASK) == MODE_STN_MONO_SINGLE) ||
	    ((jzfb->cfg & MODE_MASK) == MODE_STN_MONO_DUAL))
		pclk = pclk >> ((jzfb->cfg & STN_DAT_PINMASK) >> 4);

	if (((jzfb->cfg & MODE_MASK) == MODE_STN_COLOR_DUAL) ||
	    ((jzfb->cfg & MODE_MASK) == MODE_STN_MONO_DUAL))
		pclk >>= 1;

	return pclk;
}

static void jz_lcd_set_timing(uint32_t pclk)
{
	uint32_t val;

#ifdef CONFIG_CPU_JZ4730
        val = __cpm_get_pllout() / pclk;
        REG_CPM_CFCR2 = val - 1;
        val = pclk * 4 ;
        if ( val > 150000000 ) {
                val = 150000000;
        }
        val = __cpm_get_pllout() / val;
        val--;
        if ( val > 0xF )
                val = 0xF;
#else
	int pll_div;

	pll_div = ( REG_CPM_CPCCR & CPM_CPCCR_PCS ); /* clock source,0:pllout/2 1: pllout */
	pll_div = pll_div ? 1 : 2 ;
	val = ( __cpm_get_pllout()/pll_div ) / pclk;
	val--;
	if ( val > 0x1ff ) {
		val = 0x1ff;
	}
	__cpm_set_pixdiv(val);

	val = pclk * 3 ;	/* LCDClock > 2.5*Pixclock */
	if ( val > 150000000 ) {
		val = 150000000;
	}
	val = ( __cpm_get_pllout()/pll_div ) / val;
	val--;
	if ( val > 0x1f ) {
		val = 0x1f;
	}
#endif
	__cpm_set_ldiv( val );
	REG_CPM_CPCCR |= CPM_CPCCR_CE ; /* update divide */
}

static int jz_lcd_hw_init(vidinfo_t *vid)
{
	struct jzfb_info *jzfb = vid->jz_fb;
	struct jz_mem_info *fbi = &vid->jz_mem;
	uint32_t val = 0;
	uint32_t pclk = jz_lcd_get_pixel_clock(vid);

	/* Setting Control register */
	switch (jzfb->bpp) {
	case 1:
		val |= LCD_CTRL_BPP_1;
		break;
	case 2:
		val |= LCD_CTRL_BPP_2;
		break;
	case 4:
		val |= LCD_CTRL_BPP_4;
		break;
	case 8:
		val |= LCD_CTRL_BPP_8;
		break;
	case 15:
		val |= LCD_CTRL_RGB555;
	case 16:
		val |= LCD_CTRL_BPP_16;
		break;
#ifndef CONFIG_CPU_JZ4730
	case 17 ... 32:
		val |= LCD_CTRL_BPP_18_24;	/* target is 4bytes/pixel */
		break;
#endif
	default:
		/* printf("jz_lcd.c The BPP %d is not supported\n", jzfb->bpp); */
		val |= LCD_CTRL_BPP_16;
		break;
	}

	switch (jzfb->cfg & MODE_MASK) {
	case MODE_STN_MONO_DUAL:
	case MODE_STN_COLOR_DUAL:
		val |= jz_lcd_stn_init(jzfb->h >> 1, vid);
		break;

	case MODE_STN_MONO_SINGLE:
	case MODE_STN_COLOR_SINGLE:
		val |= jz_lcd_stn_init(jzfb->h, vid);
		break;

	case MODE_TFT_GEN:
	case MODE_TFT_CASIO:
	case MODE_8BIT_SERIAL_TFT:
	case MODE_TFT_18BIT:
		jz_lcd_tft_init(vid);
		break;

	case MODE_TFT_SAMSUNG:
	{
		jz_lcd_samsung_init(pclk, vid);
		break;
	}

	case MODE_TFT_SHARP:
	{
		jz_lcd_sharp_init(vid);
		break;
	}

	default:
		break;
	}

	/* Configure the LCD panel */

	val |= LCD_CTRL_BST_16;		/* Burst Length is 16WORD=64Byte */
	val |= LCD_CTRL_OFUP;		/* OutFIFO underrun protect */
	REG_LCD_CTRL = val;
	REG_LCD_CFG = jzfb->cfg;

	/* Timing reset. */

	__cpm_stop_lcd();
	jz_lcd_set_timing(pclk);
	__cpm_start_lcd();
	udelay(1000);

	/* Configure DMA. */

	REG_LCD_DA0 = (uint32_t) fbi->fdadr0; /* frame descriptor */

	if (((jzfb->cfg & MODE_MASK) == MODE_STN_COLOR_DUAL) ||
	    ((jzfb->cfg & MODE_MASK) == MODE_STN_MONO_DUAL))
		REG_LCD_DA1 = (uint32_t) fbi->fdadr1; /* frame descriptor */

	return 0;
}

/* Public operations. */

void lcd_set_bpp(uint8_t bpp)
{
	vidinfo_t *vid = &panel_info;
	struct jzfb_info *jzfb = vid->jz_fb;
	jzfb->bpp = bpp;
}

void lcd_enable()
{
	/* Clear the disable bit and set the enable bit. */

	REG_LCD_CTRL &= ~(1<<4);	/* LCDCTRL.DIS */
	REG_LCD_CTRL |= 1<<3;		/* LCDCTRL.ENA*/
}

void lcd_disable()
{
	REG_LCD_CTRL |= (1<<4);		/* LCDCTRL.DIS, regular disable */
}

void lcd_quick_disable()
{
	REG_LCD_CTRL &= ~(1<<4);	/* LCDCTRL.ENA, quick disable */
}

static inline uint16_t rgb8_to_rgb16(uint8_t rgb)
{
	return ((((rgb & 0xe0) >> 5) * 4) << 11) | ((((rgb & 0x1c) >> 2) * 9) << 6) | ((rgb & 0x03) * 10);
}

static inline uint16_t rgb4_to_rgb16(uint8_t rgb)
{
	return ((((rgb & 8) >> 3) * 0x1f) << 11) | ((((rgb & 6) >> 1) * 0x15) << 5) | ((rgb & 1) * 0x1f);
}

static void lcd_init_palette(vidinfo_t *vid)
{
	uint16_t *palette = (uint16_t *) lcd_get_palette(get_memory_size(), vid);
	uint16_t *end = (uint16_t *) palette + (1 << (vid->jz_fb->bpp));
	uint8_t value = 0;

	while (palette < end)
	{
		switch (vid->jz_fb->bpp)
		{
			case 4:
			*palette = rgb4_to_rgb16(value);
			break;

			case 8:
			default:
			*palette = rgb8_to_rgb16(value);
			break;
		}

		value++;
		palette++;
	}
}

uint32_t lcd_ctrl_init()
{
	vidinfo_t *vid = &panel_info;
	struct jz_mem_info *fbi = &vid->jz_mem;

	/* Start from the top of memory and obtain palette and framebuffer regions. */

	fbi->screen = lcd_get_framebuffer(get_memory_size(), 0, vid);
	fbi->palette = lcd_get_palette(get_memory_size(), vid);

	if (vid->jz_fb->bpp < 12)
		lcd_init_palette(vid);

	jz_lcd_desc_init(vid);
	jz_lcd_hw_init(vid);

	return fbi->screen;
}
